/**
 *  Source code for the ParallelSimulation class
 *  Copyright (C) 2014  Jamie Fairbrother
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <vector>
#include <cmath>
#include <thread>
#include <random>

#include "ParallelSimulation.hpp"

static const bool dbg = false;

ParallelSimulation::ParallelSimulation(std::function<double(std::mt19937 &)> dist,
				       size_t num_threads,
				       bool verbose)
  :
  SimulationBase(dist, verbose),
  num_threads(num_threads),
  mtx(),
  stop(false)
{

  if (dbg) std::cout << "ParallelSimulation: construction successful."
		     << std::endl;
}

ParallelSimulation::ParallelSimulation(const ParallelSimulation& sim)
  : SimulationBase(sim.dist, sim.verbose),
    num_threads(sim.num_threads),
    mtx(),
    stop(false)
{
  if (dbg) std::cout << "Copy construction successful." << std::endl;
}

void ParallelSimulation::accumulate()
{
  std::vector<std::thread> threads;
  for (size_t i = 0; i < num_threads; ++i)
    threads.push_back(std::thread(&ParallelSimulation::thread_accumulate, 
				  this, i));

  for (size_t i = 0; i < num_threads; ++i)
    threads[i].join();
}

void ParallelSimulation::thread_accumulate(int i) {
  double old_mean;
  double thread_mean=0.0, thread_var=0.0, thread_old_mean, thread_om2, x;
  int m = 0; // Thread counter
  std::mt19937 rng;
  rng.seed(i+1);
  while (true)
    {
      thread_old_mean=thread_mean;
      // Simulate
      x = dist(rng);

      // Update counter, and thread_mean
      ++m;
      thread_mean += (x - thread_mean)/m;
      thread_om2 = thread_old_mean*thread_old_mean;
      thread_var += thread_om2-thread_mean*thread_mean+(x*x - thread_var - thread_om2)/(double)m;

      // Print message and stop check
      if (m % batch_size == 0) 
	{
	  mtx.lock();
	  old_mean=mean;
	  mean += ((double)m/(m+n))*(thread_mean - mean);
	  var += thread_var - mean*mean - ((double)n/(n+m))*(thread_var - old_mean*old_mean) - ((double)m/(n+m))*(var - thread_mean*thread_mean);

	  n += m;
	  if (verbose) message();
	  if (stop || check_complete()) {
	    stop=true;
	    mtx.unlock();
	    break;
	  }
	  else {
	    mtx.unlock();
	    thread_mean=0.0;
	    thread_var=0.0;
	    m=0;
	  }
	}
    }
}
