/**
 *  Header file for ParallelSimulation class
 *  Copyright (C) 2014  Jamie Fairbrother
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef PARALLEL_SIMULATION_HPP
#define PARALLEL_SIMULATION_HPP

#include <functional>
#include <mutex>
#include "SimulationBase.hpp"

/**
 * Class for running a simulation where
 * the draws from the distribution are parallelized
 */

class ParallelSimulation : public SimulationBase
{
public:
  /**
   * \brief Construct instance of simulation.
   * \param dist Any function which transforms pseudo-random number 
   *             generator and returns a value.
   *             Typically this will use a distribution class from <random>
   * \param num_threads number of threads to use to complete simulation
   * \param verbose Whether simulation information should be printed 
   *        to standard output.
   */
  ParallelSimulation(std::function<double(std::mt19937 &)> dist,
		     size_t num_threads,
		     bool verbose = false);
  /**
   * \brief Destroys ParallelSimulation instance.
   */
  virtual ~ParallelSimulation() {}

  /**
   * \brief Copy constructor copies simulation parameters but simulation is reset.
   */ 
  ParallelSimulation(const ParallelSimulation&);


private:
  ParallelSimulation();           // Forbid default constructor
  void thread_accumulate(int i);
  void accumulate();
  size_t num_threads;
  std::mutex mtx;
  bool stop;
};

#endif
