/**
 *  Source code for the SimulationBase class
 *  Copyright (C) 2014  Jamie Fairbrother
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <vector>
#include <cmath>
#include <ctime>

#include <gsl/gsl_cdf.h>
#include "SimulationBase.hpp"

static const bool dbg = false;

SimulationBase::SimulationBase(std::function<double(std::mt19937&)> dist,
			       bool verbose)
  : dist(dist), mean(0.0), var(0.0), n(0), 
    verbose(verbose), start()
{}

bool SimulationBase::check_complete() {
  double scale = gsl_cdf_ugaussian_Pinv((level+1.0)/2.0);
  if (get_relative_error() *  scale >= eps_rel
      && get_standard_error() * scale >= eps_abs)
    return false;
  return true;
}

void SimulationBase::run() {
  start = time(NULL);
  char start_time[80];
  struct tm * timeinfo = localtime(&start);
  strftime(start_time, 80, "%Y-%m-%d %H:%M", timeinfo);
  if (verbose) {
    std::cerr << "\n----------------------------------" << std::endl;
    std::cerr << "----- Simulation information -----" << std::endl;
    std::cerr << "----------------------------------" << std::endl;
    std::cerr << "Required absolute precision: " << eps_abs << std::endl;
    std::cerr << "Required relative precision: " << eps_rel << std::endl;
    std::cerr << "Level: " << level << std::endl;
    std::cerr << "Simulation started on: " << start_time << std::endl;
    std::cerr << "----------------------------------\n" << std::endl;
  }

  accumulate();

  if (verbose) {
    std::cout << "\nSimulation successfully completed after "
	      << std::fixed << std::setprecision(1)
	      << difftime(time(NULL),start) << " seconds." <<std::endl;
  }
}

double SimulationBase::get_mean()
{
  return mean;
}

double SimulationBase::get_standard_deviation()
{
  return sqrt((double(n)/(n-1))*var);
}

double SimulationBase::get_standard_error()
{
  return get_standard_deviation()/sqrt(n);
}

double SimulationBase::get_relative_error()
{
  return get_standard_error()/std::abs(get_mean());
}

void SimulationBase::message()
{
  std::cout << "\nAfter " << n << " samples..." << std::endl;
  std::cout << "Mean:" << get_mean();
  std::cout << "\nStandard error: " << std::setprecision(4) 
	    << get_standard_error();
  std::cout << "\nRelative error: " << std::setprecision(4)
	    << get_relative_error();
  std::cout << std::endl;
}

SimulationBase& SimulationBase::set_eps_abs(double eps_abs) {
  if (eps_abs < 0.0) throw std::invalid_argument("set_eps_abs: eps_abs must be positive");
  if (eps_abs == 0.0 && this->eps_rel == 0.0) throw std::invalid_argument("set_eps_abs: both eps_abs and eps_rel cannot be zero"); 
  this->eps_abs = eps_abs;
  return *this;
}

SimulationBase& SimulationBase::set_eps_rel(double eps_rel) {
  if (eps_rel < 0.0) throw std::invalid_argument("set_eps_rel: eps_rel must be positive");
  if (eps_abs == 0.0 && this->eps_rel == 0.0) throw std::invalid_argument("set_eps_rel: both eps_abs and eps_rel cannot be zero"); 
  this->eps_rel = eps_rel;
  return *this;
}

SimulationBase& SimulationBase::set_batch_size(int batch_size) {
  if (batch_size <= 0) throw std::invalid_argument("set_batch_size: batch_size must be a strictly postive integer");
  this->batch_size = batch_size;
  return *this;
}

SimulationBase& SimulationBase::set_level(double level) {
  if (level <= 0 || level >= 1) throw std::invalid_argument("set_level: level must be a strictly between 0 and 1");
  this->level = level;
  return *this;
}
