/**
 *  Header file for SimulationBase class
 *  Copyright (C) 2014  Jamie Fairbrother
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIMULATION_BASE_HPP
#define SIMULATION_BASE_HPP

#include <functional>
#include <random>

/**
 * Provides a general framework for running Monte Carlo
 * simulation
 */

class SimulationBase
{
public:
  /**
   * \brief Construct instance of simulation.
   * \param rng Random number generator.
   * \param verbose Whether simulation information should be printed to standard output.
   */
  SimulationBase(std::function< double(std::mt19937 &) > dist, 
		 bool verbose = false);
  
  /**
   * \brief Destroys SimpleSimulation instance.
   */
  virtual ~SimulationBase() {}

  /**
   * \brief Run the simulation.
   */
  void run();

  /**
   * \brief Prints message giving information on the current state of the simulation.
   */
  virtual void message();

  /**
   * \brief Check whether sample mean in the simulation has converged to the
   *        specified relative precision.
   *         
   *        Specifically, this checks whether the standard error 
   *        and relative error of the sample is less than eps_abs 
   *        or eps_rel respectively.
   *        
   * \return bool value specifying whether or not simulation is complete.
   */
  bool check_complete();

  /**
   * \brief Get current mean.
   */
  double get_mean();

  /**
   * \brief Get current standard deviation.
   */
  double get_standard_deviation();

  /**
   * \brief Get current standard error.
   */
  double get_standard_error();


  /**
   * \brief Get relative error.
   * The relative error is defined to be the standard error of a random
   * variable divided by its mean.
   */
  double get_relative_error();

  /**
   * \brief Set absolute precision for simulation
   *
   * eps_abs must be positive
   */
  SimulationBase &set_eps_abs(double eps_abs);

  /**
   * \brief Set relative precision for simulation
   *
   * eps_rel must be positive
   */
  SimulationBase &set_eps_rel(double eps_rel);

  /**
   * \brief Set batch_size for simulation
   *
   * The batch size of the simulation is every
   * how many samples the stopping condition is
   * checked.
   */
  SimulationBase &set_batch_size(int batch_size);

  /**
   * \brief Set level for simulation
   * 
   * The level corresponds to the probability
   * that the sample mean has been calculated to
   * the accuracy specified by eps_abs and eps_rel.
   */
  SimulationBase &set_level(double level);

protected:
  // A distribution is just a function of the Mersenne twister PRNG
  std::function< double(std::mt19937 &) > dist;
  int batch_size = 100000;
  double mean;
  double var;
  int n; // Sample Counter
  bool verbose;
private:

  /*
   * The main part of the simulation. Must be defined in 
   * derived classes.
   */
  virtual void accumulate() = 0;

  double eps_abs = 1e-4; /// Required absolute precision
  double eps_rel = 1e-2; /// Required relative precision
  double level = 0.95;   /// level of confidence interval for stop test
  time_t start;          /// time when simulation is started
};

#endif
